@@ -1,6 +1,6 @@
 // 
 //
-//  $Id: spamass-milter.cpp,v 1.13 2002/08/14 16:15:10 dnelson Exp $
+//  $Id: spamass-milter.cpp,v 1.22 2002/12/31 20:03:13 dnelson Exp $
 //
 //  SpamAss-Milter 
 //    - a rather trivial SpamAssassin Sendmail Milter plugin
@@ -57,7 +57,7 @@
 //   7. check for the flags affected by SpamAssassin and set/change
 //      them accordingly
 //   8. replace the body with the one provided by SpamAssassin if the
-//      mail was rated spam.
+//      mail was rated spam, unless -m is specified
 //   9. free all temporary data
 //   10. tell sendmail to let the mail to go on (default) or be discarded
 //    -- wait for mail to show up -- (restart at 3)
@@ -104,6 +104,8 @@ extern "C" {
 
 // }}} 
 
+static const char Id[] = "$Id: spamass-milter.cpp,v 1.22 2002/12/31 20:03:13 dnelson Exp $";
+
 struct smfiDesc smfilter =
   {
     "SpamAssassin", // filter name
@@ -111,8 +113,8 @@ struct smfiDesc smfilter =
     SMFIF_ADDHDRS|SMFIF_CHGHDRS|SMFIF_CHGBODY,  // flags
     NULL, // info filter callback
     NULL, // HELO filter callback
-    mlfi_envfrom, // envelope filter callback
-    NULL, // envelope recipient filter callback
+    mlfi_envfrom, // envelope sender filter callback
+    mlfi_envrcpt, // envelope recipient filter callback
     mlfi_header, // header filter callback
     mlfi_eoh, // end of header callback
     mlfi_body, // body filter callback
@@ -121,7 +123,12 @@ struct smfiDesc smfilter =
     mlfi_close, // connection cleanup callback
   };
 
-int flag_debug=0;
+int flag_debug = 0;
+bool flag_reject = false;
+bool flag_sniffuser = false;
+int reject_score = -1;
+bool dontmodify = false;
+char *defaultuser;
 
 // {{{ main()
 
@@ -129,7 +136,7 @@ int
 main(int argc, char* argv[])
 {
    int c, err = 0;
-   const char *args = "p:fd:";
+   const char *args = "p:fd:mr:u:";
    char *sock = NULL;
    bool dofork = false;
 
@@ -145,6 +152,17 @@ main(int argc, char* argv[])
 			case 'd':
 				flag_debug = atoi(optarg);
 				break;
+			case 'm':
+				dontmodify = true;
+				break;
+			case 'r':
+				flag_reject = true;
+				reject_score = atoi(optarg);
+				break;
+			case 'u':
+				flag_sniffuser = true;
+				defaultuser = strdup(optarg);
+				break;
 			case '?':
 				err = 1;
 				break;
@@ -154,10 +172,15 @@ main(int argc, char* argv[])
    if (!sock || err) {
       cout << PACKAGE_NAME << " - Version " << PACKAGE_VERSION << endl;
       cout << "SpamAssassin Sendmail Milter Plugin" << endl;
-      cout << "Usage: spamass-milter -p socket [-f] [-d nn]" << endl;
+      cout << "Usage: spamass-milter -p socket [-d nn] [-f] [-m] [-r nn] [-u user]" << endl;
       cout << "   -p socket: path to create socket" << endl;
+      cout << "   -d nn: set debug level to nn (1-3).  Logs to syslog" << endl;
       cout << "   -f: fork into background" << endl;
-      cout << "   -d nn: set debug level to nn (1 or 2).  Logs to syslog" << endl;
+      cout << "   -m: don't modify body, Content-type: or Subject:" << endl;
+      cout << "   -r nn: reject messages with a score >= nn with an SMTP error.\n" 
+              "          use -1 to reject any messages tagged by SA." << endl;
+      cout << "   -u user: pass the recipient's username to spamc.  Use 'user'\n"
+              "          if there are multiple recipients." << endl;
       exit(EX_USAGE);
    }
 
@@ -187,21 +210,60 @@ main(int argc, char* argv[])
 	} else {
       debug(1, "smfi_register succeeded");
    }
-	return smfi_main();
+	debug(0, "spamass-milter %s starting", PACKAGE_VERSION);
+	err = smfi_main();
+	debug(0, "spamass-milter %s exiting", PACKAGE_VERSION);
+	return err;
 };
 
 // }}}
 
+/* Update a header if SA changes it, or add it if it is new. */
+void update_or_insert(SpamAssassin* assassin, SMFICTX* ctx, string oldstring, t_setter setter, char *header )
+{
+	string::size_type eoh1(assassin->d().find("\n\n"));
+	string::size_type eoh2(assassin->d().find("\n\r\n"));
+	string::size_type eoh = ( eoh1 < eoh2 ? eoh1 : eoh2 );
+
+	string newstring;
+	string::size_type oldsize;
+
+	debug(3, "u_or_i: looking at <%s>", header);
+	debug(3, "u_or_i: oldstring: <%s>", oldstring.c_str());
+
+	newstring = retrieve_field(assassin->d().substr(0, eoh), string(header));
+	debug(3, "u_or_i: newstring: <%s>", newstring.c_str());
+
+	oldsize = callsetter(*assassin,setter)(newstring);
+      
+	if (newstring != oldstring)
+	{
+		/* change if old one was present, append if non-null */
+		if (oldsize > 0)
+		{
+			debug(3, "u_or_i: changing");
+			smfi_chgheader(ctx, header, 1, newstring.size() > 0 ? 
+				const_cast<char*>(newstring.c_str()) : NULL );
+		} else if (newstring.size() > 0)
+		{
+			debug(3, "u_or_i: inserting");
+			smfi_addheader(ctx, header, 
+				const_cast<char*>(newstring.c_str()));
+		}
+	} else
+	{
+		debug(3, "u_or_i: no change");
+	}
+}
+
 // {{{ Assassinate
 
 //
-// implement the changes suggested by SpamAssassin for the mail.
-//
-void
+// implement the changes suggested by SpamAssassin for the mail.  Returns
+// the milter error code.
+int 
 assassinate(SMFICTX* ctx, SpamAssassin* assassin)
 {
-  string::size_type old;
-
   // find end of header (eol in last line of header)
   // and beginning of body
   string::size_type eoh1(assassin->d().find("\n\n"));
@@ -209,89 +271,42 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
   string::size_type eoh = ( eoh1 < eoh2 ? eoh1 : eoh2 );
   string::size_type bob = assassin->d().find_first_not_of("\r\n", eoh);
 
+  update_or_insert(assassin, ctx, assassin->spam_flag(), &SpamAssassin::set_spam_flag, "X-Spam-Flag");
+  update_or_insert(assassin, ctx, assassin->spam_status(), &SpamAssassin::set_spam_status, "X-Spam-Status");
 
-  // X-Spam-Status header //
-  // find it:
-  old = assassin->set_spam_status(retrieve_field(assassin->d().substr(0, eoh), 
-						 string("X-Spam-Status")));
-
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Status",1,assassin->spam_status().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_status().c_str()) : NULL );
-  else if (assassin->spam_status().size()>0)
-      smfi_addheader(ctx, "X-Spam-Status", 
-		     const_cast<char*>(assassin->spam_status().c_str()));
-
-
-  // X-Spam-Flag header //
-  // find it:
-  old = assassin->set_spam_flag(retrieve_field(assassin->d().substr(0, eoh), string("X-Spam-Flag")));
-
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Flag",1,assassin->spam_flag().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_flag().c_str()) : NULL );
-  else if (assassin->spam_flag().size()>0)
-      smfi_addheader(ctx, "X-Spam-Flag", 
-		     const_cast<char*>(assassin->spam_flag().c_str()));
-  
-
-  // X-Spam-Report header //
-  // find it:
-  old = assassin->set_spam_report(retrieve_field(assassin->d().substr(0, eoh), 
-						 string("X-Spam-Report")));
-    
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Report",1,assassin->spam_report().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_report().c_str()) : NULL );
-  else if (assassin->spam_report().size()>0)
-	smfi_addheader(ctx, "X-Spam-Report", 
-		       const_cast<char*>(assassin->spam_report().c_str()));
-  
-
-  // X-Spam-Prev-Content-Type header //
-  // find it:
-  old = assassin->set_spam_prev_content_type(retrieve_field(assassin->d().substr(0, eoh), 
-							    string("X-Spam-Prev-Content-Type")));
-  
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Prev-Content-Type",1,assassin->spam_prev_content_type().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_prev_content_type().c_str()) : NULL );
-  else if (assassin->spam_prev_content_type().size()>0)
-    smfi_addheader(ctx, "X-Spam-Prev-Content-Type", 
-		   const_cast<char*>(assassin->spam_prev_content_type().c_str()));
-
-
-  // X-Spam-Level header //
-  // find it:
-  old = assassin->set_spam_level(retrieve_field(assassin->d().substr(0, eoh), 
-						string("X-Spam-Level")));
-  
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Level",1,assassin->spam_level().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_level().c_str()) : NULL );
-  else if (assassin->spam_level().size()>0)
-    smfi_addheader(ctx, "X-Spam-Level", 
-		   const_cast<char*>(assassin->spam_level().c_str()));
-  
-
-  // X-Spam-Checker-Version header //
-  // find it:
-  old = assassin->set_spam_checker_version(retrieve_field(assassin->d().substr(0, eoh), 
-							  string("X-Spam-Checker-Version")));
-  
-  // change if old one was present, append if non-null
-  if (old > 0)
-    smfi_chgheader(ctx,"X-Spam-Checker-Version",1,assassin->spam_checker_version().size() > 0 ? 
-		   const_cast<char*>(assassin->spam_checker_version().c_str()) : NULL );
-  else if (assassin->spam_checker_version().size()>0)
-    smfi_addheader(ctx, "X-Spam-Checker-Version", 
-		   const_cast<char*>(assassin->spam_checker_version().c_str()));
+  /* Summarily reject the message if SA tagged it, or if we have a minimum
+     score, reject if it exceeds that score. */
+  if (flag_reject)
+  {
+	bool do_reject = false;
+	if (reject_score == -1 && assassin->spam_flag().size()>0)
+		do_reject = true;
+	if (reject_score != -1)
+	{
+		int score, rv;
+		const char *spam_status = assassin->spam_status().c_str();
+		rv = sscanf(spam_status,"%*s hits=%d", &score);
+		if (rv != 1)
+			debug(0, "Could not extract score from <%s>", spam_status);
+		else 
+		{
+			debug(1, "SA score: %d", score);
+			if (score >= reject_score)
+				do_reject = true;
+		}
+	}
+	if (do_reject)
+	{
+		debug(1, "Rejecting");
+		smfi_setreply(ctx, "550", "5.7.1", "Blocked by SpamAssassin");
+		return SMFIS_REJECT;
+	}
+  }
 
+  update_or_insert(assassin, ctx, assassin->spam_report(), &SpamAssassin::set_spam_report, "X-Spam-Report");
+  update_or_insert(assassin, ctx, assassin->spam_prev_content_type(), &SpamAssassin::set_spam_prev_content_type, "X-Spam-Prev-Content-Type");
+  update_or_insert(assassin, ctx, assassin->spam_level(), &SpamAssassin::set_spam_level, "X-Spam-Level");
+  update_or_insert(assassin, ctx, assassin->spam_checker_version(), &SpamAssassin::set_spam_checker_version, "X-Spam-Checker-Version");
 
   // 
   // If SpamAssassin thinks it is spam, replace
@@ -302,55 +317,20 @@ assassinate(SMFICTX* ctx, SpamAssassin* assassin)
   //  However, only issue the header replacement calls if the content has
   //  actually changed. If SA didn't change subject or content-type, don't
   //  replace here unnecessarily.
-  if (assassin->spam_flag().size()>0)
+  if (!dontmodify && assassin->spam_flag().size()>0)
     {
-      string oldstring;
-      string newstring;
+	  update_or_insert(assassin, ctx, assassin->subject(), &SpamAssassin::set_subject, "Subject");
+	  update_or_insert(assassin, ctx, assassin->content_type(), &SpamAssassin::set_content_type, "Content-Type");
 
-      // Subject header //
-      // find it:
-      newstring = retrieve_field(assassin->d().substr(0, eoh), string("Subject"));
-      oldstring = assassin->subject();
-
-      old = assassin->set_subject(oldstring);
-      
-      // change if old one was present, append if non-null
-      if (old > 0 && newstring != oldstring)
-	smfi_chgheader(ctx,"Subject",1,newstring.size() > 0 ? 
-		       const_cast<char*>(newstring.c_str()) : NULL );
-      else if (newstring.size()>0)
-	smfi_addheader(ctx, "Subject", 
-		       const_cast<char*>(newstring.c_str()));
-      
-
-      // Content-Type header //
-      // find it:
-      newstring = retrieve_field(assassin->d().substr(0, eoh), string("Content-Type"));
-      oldstring = assassin->content_type();
-
-      old = assassin->set_content_type(oldstring);
-      
-      // change if old one was present, append if non-null
-      if (old > 0 && newstring != oldstring)
-	smfi_chgheader(ctx,"Content-Type",1,newstring.size() > 0 ? 
-		       const_cast<char*>(newstring.c_str()) : NULL );
-      else if (assassin->content_type().size()>0)
-	smfi_addheader(ctx, "Content-Type", 
-		       const_cast<char*>(newstring.c_str()));
-      
-      
-      // Replace body with the one SpamAssassin provided //
+      // Replace body with the one SpamAssassin provided
       string::size_type body_size = assassin->d().size() - bob;
-      unsigned char* bodyp = (unsigned char*) 
-	const_cast<char*>(assassin->d().substr(bob, string::npos).c_str());
-      if ( smfi_replacebody(ctx, bodyp, body_size) == MI_FAILURE )
+      string body=assassin->d().substr(bob, string::npos);
+      if ( smfi_replacebody(ctx, (unsigned char *)body.c_str(), body_size) == MI_FAILURE )
 	throw string("error. could not replace body.");
       
     };
 
-  // erase mail right away
-  assassin->d().erase();
-  
+  return SMFIS_CONTINUE;
 };
 
 // retrieve the content of a specific field in the header
@@ -363,13 +343,21 @@ retrieve_field(const string& header, const string& field)
 
   // return empty string if not found
   if (pos == string::npos)
+  {
+    debug(3, "r_f: failed");
     return string("");
+  }
 
   // look for end of field name
   pos = find_nocase(header, string(" "), pos) + 1;
   
-  // look for end of content
   string::size_type pos2(pos);
+
+  // is field empty? 
+  if (pos2 == find_nocase(header, string("\n"), pos2))
+    return string("");
+
+  // look for end of content
   do {
 
     pos2 = find_nocase(header, string("\n"), pos2+1);
@@ -416,6 +404,28 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
   return SMFIS_CONTINUE;
 };
 
+//
+// Gets called once for each recipient
+//
+// stores the first recipient in the spamassassin object and
+// discards the rest, keeping track of the number of recipients.
+//
+sfsistat
+mlfi_envrcpt(SMFICTX* ctx, char** envrcpt)
+{
+	SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
+
+	if (assassin->numrcpt() == 0)
+	{
+		assassin->set_numrcpt(1);
+		assassin->set_rcpt(string(envrcpt[0]));
+	} else
+	{
+		assassin->set_numrcpt();
+	}
+	return SMFIS_CONTINUE;
+}
+
 //
 // Gets called repeatedly for all header fields
 //
@@ -425,41 +435,59 @@ mlfi_envfrom(SMFICTX* ctx, char** envfrom)
 // only exception: SpamAssassin header fields (X-Spam-*) get suppressed
 // but are being stored in the SpamAssassin element.
 //
+// this function also starts the connection with the SPAMC program the
+// first time it is called.
+//
+
 sfsistat
 mlfi_header(SMFICTX* ctx, char* headerf, char* headerv)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
   debug(1, "mlfi_header: enter");
 
+  // Check if the SPAMC program has already been run, if not we run it.
+  if ( !(assassin->connected) )
+     {
+       try {
+         assassin->connected = 1; // SPAMC is getting ready to run
+         assassin->Connect();
+       } 
+       catch (string& problem) {
+         throw_error(problem);
+         return SMFIS_TEMPFAIL;
+       };
+     }
+ 
+
   // Is it a "X-Spam-" header field?
   if ( cmp_nocase_partial(string("X-Spam-"), string(headerf)) == 0 )
     {
+      int suppress = 1;
       // memorize content of old fields
 
-      // X-Spam-Status:
       if ( cmp_nocase_partial(string("X-Spam-Status"), string(headerf)) == 0 )
 	assassin->set_spam_status(string(headerv));
-      
-      // X-Spam-Flag:
-      if ( cmp_nocase_partial(string("X-Spam-Flag"), string(headerf)) == 0 )
+      else if ( cmp_nocase_partial(string("X-Spam-Flag"), string(headerf)) == 0 )
 	assassin->set_spam_flag(string(headerv));
-      
-      // X-Spam-Report:
-      if ( cmp_nocase_partial(string("X-Spam-Report"), string(headerf)) == 0 )
+      else if ( cmp_nocase_partial(string("X-Spam-Report"), string(headerf)) == 0 )
 	assassin->set_spam_report(string(headerv));
-
-      // X-Spam-Prev-Content-Type:
-      if ( cmp_nocase_partial(string("X-Spam-Prev-Content-Type"), string(headerf)) == 0 )
+      else if ( cmp_nocase_partial(string("X-Spam-Prev-Content-Type"), string(headerf)) == 0 )
 	assassin->set_spam_prev_content_type(string(headerv));
-
-      // X-Spam-Level:
-      if ( cmp_nocase_partial(string("X-Spam-Level"), string(headerf)) == 0 )
+      else if ( cmp_nocase_partial(string("X-Spam-Level"), string(headerf)) == 0 )
 	assassin->set_spam_level(string(headerv));
-      
-      // X-Spam-Checker-Version:
-      if ( cmp_nocase_partial(string("X-Spam-Checker-Version"), string(headerf)) == 0 )
+      else if ( cmp_nocase_partial(string("X-Spam-Checker-Version"), string(headerf)) == 0 )
 	assassin->set_spam_checker_version(string(headerv));
-
+      else
+      {
+      	/* Hm. X-Spam header, but not one we recognize.  Pass it through. */
+      	suppress = 0;
+      }
+      
+      if (suppress)
+      {
+	debug(1, "mlfi_header: suppress");
+	return SMFIS_CONTINUE;
+      }
     };
 
   // Content-Type: will be stored if present
@@ -505,6 +533,19 @@ mlfi_eoh(SMFICTX* ctx)
 
   debug(1, "mlfi_eoh: enter");
 
+  // Check if the SPAMC program has already been run, if not we run it.
+  if ( !(assassin->connected) )
+     {
+       try {
+         assassin->connected = 1; // SPAMC is getting ready to run
+         assassin->Connect();
+       } 
+       catch (string& problem) {
+         throw_error(problem);
+         return SMFIS_TEMPFAIL;
+       };
+     }
+
   try {
     // add blank line between header and body
     assassin->output("\n\n",2);
@@ -562,6 +603,7 @@ sfsistat
 mlfi_eom(SMFICTX* ctx)
 {
   SpamAssassin* assassin = static_cast<SpamAssassin*>(smfi_getpriv(ctx));
+  int milter_status;
  
   debug(1, "mlfi_eom: enter");
   try {
@@ -572,12 +614,7 @@ mlfi_eom(SMFICTX* ctx)
     // read what the Assassin is telling us
     assassin->input();
 
-    // It makes no sense to modify the mail if it already rated
-    // Spam. Otherwise this is our chance to modify the mail
-    // accordingly to what the SpamAssassin told us.
-    //
-    if (assassin->spam_flag().size()==0)
-      assassinate(ctx, assassin);
+    milter_status = assassinate(ctx, assassin);
 
     // now cleanup the element.
     smfi_setpriv(ctx, static_cast<void*>(0));
@@ -594,7 +631,7 @@ mlfi_eom(SMFICTX* ctx)
   
   // go on...
   debug(1, "mlfi_eom: exit");
-  return SMFIS_CONTINUE;
+  return milter_status;
 };
 
 //
@@ -629,9 +666,49 @@ mlfi_abort(SMFICTX* ctx)
 
 // {{{ SpamAssassin Class
 
+//
+// This is a new constructor for the SpamAssassin object.  It simply 
+// initializes two variables.  The original constructor has been
+// renamed to Connect().
+//
 SpamAssassin::SpamAssassin():
-  error(false)
+  error(false),
+  connected(false),
+  _numrcpt(0)
+{
+}
+
+
+SpamAssassin::~SpamAssassin()
 { 
+	if (connected) 
+	{
+		// close all pipes that are still open
+		if (pipe_io[0][0] > -1)	close(pipe_io[0][0]);
+		if (pipe_io[0][1] > -1)	close(pipe_io[0][1]);
+		if (pipe_io[1][0] > -1)	close(pipe_io[1][0]);
+		if (pipe_io[1][1] > -1)	close(pipe_io[1][1]);
+
+		// child still running?
+		if (running)
+		{
+			// slaughter child
+			kill(pid, SIGKILL);
+
+			// wait for child to terminate
+			int status;
+			waitpid(pid, &status, 0);
+		}
+	}
+};
+
+//
+// This is the old SpamAssassin constructor.  It has been renamed Connect(),
+// and is now called at the beginning of the mlfi_header() function.
+//
+
+void SpamAssassin::Connect()
+{
   // set up pipes for in- and output
   if (pipe(pipe_io[0]))
     throw string(string("pipe error: ")+string(strerror(errno)));
@@ -664,16 +741,32 @@ SpamAssassin::SpamAssassin():
       // execute spamc 
       // absolute path (determined in autoconf) 
       // should be a little more secure
+      // XXX arbitrary 100-argument max
+      int argc = 0;
+      char** argv = (char**) malloc(100*sizeof(char*));
 #if 0
-      char** argv = (char**) malloc(4*sizeof(char*));
-      argv[0] = SPAMC;
-      argv[1] = "-u";
-      argv[2] = "milter";
-      argv[3] = 0;
+      argv[argc++] = SPAMC;
+      argv[argc++] = "-u";
+      argv[argc++] = "milter";
+      argv[argc++] = 0;
 #else
-      char** argv = (char**) malloc(2*sizeof(char*));
-      argv[0] = SPAMC;
-      argv[1] = 0;
+      argv[argc++] = SPAMC;
+      if (flag_sniffuser) 
+      {
+        argv[argc++] = "-u";
+        if ( numrcpt() != 1 )
+        {
+          // More (or less?) than one recipient, so we pass the special
+          // username _multi_ to SPAMC.  This way special rules can be
+          // defined for multi recipient messages.
+          argv[argc++] = "_multi_"; 
+        } else
+        { 
+          // There is only 1 recipient so we pass the username to SPAMC 
+          argv[argc++] = (char *) local_user().c_str(); 
+        }
+      }
+      argv[argc++] = 0;
 #endif
 
       execvp(argv[0] , argv); // does not return!
@@ -682,7 +775,7 @@ SpamAssassin::SpamAssassin():
       throw_error(string("execution error: ")+string(strerror(errno)));
       
       break;
-    };
+    }
 
   // +++ PARENT +++
 
@@ -705,35 +798,6 @@ SpamAssassin::SpamAssassin():
 
 };
 
-SpamAssassin::~SpamAssassin()
-{ 
-
-  // close all pipes that are still open
-  if (pipe_io[0][0] > -1)
-    close(pipe_io[0][0]);
-
-  if (pipe_io[0][1] > -1)
-    close(pipe_io[0][1]);
-
-  if (pipe_io[1][0] > -1)
-    close(pipe_io[1][0]);
-
-  if (pipe_io[1][1] > -1)
-    close(pipe_io[1][1]);
-
-  // child still running?
-  if (running)
-    {
-      // slaughter child
-      kill(pid, SIGKILL);
-
-      // wait for child to terminate
-      int status;
-      waitpid(pid, &status, 0);
-    };
-
-};
-
 // write to SpamAssassin
 void
 SpamAssassin::output(const void* buffer, long size)
@@ -756,12 +820,12 @@ SpamAssassin::output(const void* buffer, long size)
 	fds[1].fd = pipe_io[1][0];
 	fds[1].events = POLLIN;
 
-	debug(2, "polling");
+	debug(2, "polling fds %d and %d", pipe_io[0][1], pipe_io[1][0]);
 	nready = poll(fds, nfds, 1000);
 	if (nready == -1)
 		throw("poll failed");
 
-	debug(2, "poll returned %d", nready);
+	debug(2, "poll returned %d, fd0=%d, fd1=%d", nready, fds[0].revents, fds[1].revents);
 
 	if (fds[1].revents & POLLIN)
 	{
@@ -902,6 +966,40 @@ SpamAssassin::subject()
   return _subject;
 };
 
+string&
+SpamAssassin::rcpt()
+{
+  return _rcpt;
+}
+
+string
+SpamAssassin::local_user()
+{
+  // assuming we have a recipient in the form: <username@somehost.somedomain>
+  // we return 'username'
+  return _rcpt.substr(1,_rcpt.find('@')-1);
+}
+
+int
+SpamAssassin::numrcpt()
+{
+  return _numrcpt;
+}
+
+int
+SpamAssassin::set_numrcpt()
+{
+  _numrcpt++;
+  return _numrcpt;
+}
+
+int
+SpamAssassin::set_numrcpt(const int val)
+{
+  _numrcpt = val;
+  return _numrcpt;
+}
+
 //
 // set the values of the different SpamAssassin
 // fields in our element. Returns former size of field
@@ -970,6 +1068,14 @@ SpamAssassin::set_subject(const string& val)
   return (old);
 };
 
+string::size_type
+SpamAssassin::set_rcpt(const string& val)
+{
+  string::size_type old = _rcpt.size();
+  _rcpt = val;
+  return (old);  
+}
+
 //
 // Read available output from SpamAssassin client
 //
@@ -1105,12 +1211,16 @@ find_nocase(const string& array, const string& pattern, string::size_type start)
 	{
 	  ++ctr;
 	  if (ctr == pattern.size())
+	  {
+	    debug(3, "f_nc: <%s><%s>: hit", array.c_str(), pattern.c_str());
 	    return pos;
+	  }
 	};
       
       ++pos;
     };
 
+  debug(3, "f_nc: <%s><%s>: nohit", array.c_str(), pattern.c_str());
   return string::npos;
 };
 
@@ -1123,11 +1233,15 @@ cmp_nocase_partial(const string& s, const string& s2)
 
   while ( p != s.end() && p2 != s2.end() ) {
     if (toupper(*p) != toupper(*p2))
+    {
+      debug(3, "c_nc_p: <%s><%s> : miss", s.c_str(), s2.c_str());
       return (toupper(*p) < toupper(*p2)) ? -1 : 1;
+    }
     ++p;
     ++p2;
   };
 
+  debug(3, "c_nc_p: <%s><%s> : hit", s.c_str(), s2.c_str());
   return 0;
 
 };
